from abc import ABC
from collections import OrderedDict
from xfact.registry.registrable import Registrable


class PostProcessor(Registrable, ABC):

    def __init__(self, model, tokenizer):
        self.model = model
        self.tokenizer = tokenizer

    def process_text(self, examples, features, predictions, trainer):
        raise NotImplementedError()


    def clean(self, text):
        # Check the internal _bos_token to prevent tokenizer from spamming errors
        return text.replace(self.tokenizer.pad_token, "")\
            .replace(self.tokenizer.unk_token, "<UNK>")\
            .replace(self.tokenizer.eos_token if self.tokenizer._eos_token else "", "")\
            .replace(self.tokenizer.sep_token if self.tokenizer._sep_token else "", "")\
            .replace(self.tokenizer.bos_token if self.tokenizer._bos_token else "", "")


@PostProcessor.register("default")
class DefaultPostProcessor(PostProcessor):
    def process_text(self, examples, features, predictions, trainer):
        predicted = [self.clean(p).strip() for p in features.tokenizer.batch_decode(predictions.predictions)]
        actual = [self.clean(p).strip() for p in features.tokenizer.batch_decode(predictions.label_ids)]

        assert len(actual) == len(examples) == len(predicted)

        return {
            "predicted": predicted,
            "actual": actual
        }


@PostProcessor.register("nested")
class NestedPostProcessor(PostProcessor):
    def process_text(self, examples, features, predictions, trainer):
        predicted = [
            [self.clean(r).strip() for r in p.split(self.tokenizer.sep_token)] for p in
            features.tokenizer.batch_decode(predictions.predictions)
        ]

        actual = [
            [self.clean(r).strip() for r in p.split(self.tokenizer.sep_token)] for p in
            features.tokenizer.batch_decode(predictions.label_ids)
        ]

        assert len(actual) == len(examples) == len(predicted)

        return {
            "predicted": predicted,
            "actual_flat": actual,
            "actual": [f['nested_entities'] for f in features.instances]
        }




import pickle
from typing import List, Iterable
import logging

logger = logging.getLogger(__name__)


class PrefixTree():
    def __init__(self):
        self.root = dict()

    def add_string(self, other:Iterable[int]):
        current = self.root
        for tok in other:
            previous = current
            assert previous is not None

            current = current.get(tok, None)

            if current is None:
                current = dict()
                previous[tok] = current

    def get_next(self, test_string: Iterable[int]):
        current = self.root

        for tok in test_string:
            current = current.get(tok, None)

            if current is None:
                return None

        return current


def single_document_prefix(tokenizer, path):
    with open(path, "rb") as f:
        print("loading prefix decoder")
        prefix_decoder = pickle.load(f)
        print("dne")

    def do_prefix_decode(batch_id, hypothesis) -> List[int]:
        decode_string = hypothesis.cpu().tolist()

        if len(decode_string) == 1 and (decode_string[0] == tokenizer.pad_token_id):
            return list(prefix_decoder.root.keys())

        elif len(decode_string) > 1 and decode_string[-1] == tokenizer.eos_token_id:
            return []

        if decode_string[0] == 0:
            decode_string = decode_string[1:]

        next_tokens = prefix_decoder.get_next(decode_string)

        if next_tokens is None:
            return []

        return list(next_tokens.keys())

    return do_prefix_decode


def multi_document_prefix(tokenizer, path):
    with open(path, "rb") as f:
        logger.info("Loading prefix decoder")
        prefix_decoder = pickle.load(f)
        logger.info("Done")

    def do_prefix_decode_line_number(batch_id, hypothesis) -> List[int]:
        decode_string = hypothesis.cpu().tolist()

        if decode_string[0] == 0:
            decode_string = decode_string[1:]

        if tokenizer.sep_token_id in decode_string:
            decode_string = decode_string[len(decode_string) - decode_string[::-1].index(tokenizer.sep_token_id):]

        if len(decode_string) == 0:
            return list(prefix_decoder.root.keys())

        elif len(decode_string) > 1 and decode_string[-1] == tokenizer.eos_token_id:
            return []

        next_tokens = prefix_decoder.get_next(decode_string)
        if next_tokens is None:
            return []

        if tokenizer.eos_token_id in next_tokens:
            return [tokenizer.sep_token_id] + list(next_tokens.keys())

        return list(next_tokens.keys())

    return do_prefix_decode_line_number


from setuptools import setup, find_packages

with open('requirements.txt') as f:
    reqs = f.read()

reqs = reqs.strip().split('\n')

install = [req for req in reqs if not req.startswith("git+git://")]
depends = [req.replace("git+git://", "git+http://") for req in reqs if req.startswith("git+git://")]


setup(
    name='xfact',
    version='0.0.15',
    author='James Thorne',
    author_email='james@jamesthorne.com',
    description='xfact',
    long_description="readme",
    python_requires='>=3.8',
    packages=["xfact","xfact.nlp","xfact.config","xfact.logs","xfact.registry", "xfact.nlp.inference"],
    install_requires=install,
    dependency_links=depends,
    package_dir={'':'src'}
)


from dataclasses import field, dataclass
from typing import Optional


@dataclass
class ModelArguments:
    """
    Arguments pertaining to which model/config/tokenizer we are going to fine-tune from.
    """
    model_name_or_path: str = field(
        metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models"}
    )
    predict_single:Optional[bool] = field(
        default=False,
        metadata={"help": "Limit decoder to signel output"}
    )
    prefix_path : Optional[str] = field(
        metadata={"help": "Path to TRIE file"},
        default="wikipedia-titles-structured-pt.pkl"
    )
    config_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained config name or path if not the same as model_name"}
    )
    tokenizer_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained tokenizer name or path if not the same as model_name"}
    )
    cache_dir: Optional[str] = field(
        default=None,
        metadata={"help": "Path to directory to store the pretrained models downloaded from huggingface.co"},
    )
    model_revision: str = field(
        default="main",
        metadata={"help": "The specific model version to use (can be a branch name, tag name or commit id)."},
    )
    use_auth_token: bool = field(
        default=False,
        metadata={
            "help": "Will use the token generated when running `transformers-cli login` (necessary to use this script "
            "with private models)."
        },
    )
    dropout_rate: float = field(
        default=None,
        metadata={
            "help": "Overwrite dropout rate for model training"
        }
    )


@dataclass
class DataTrainingArguments:
    """
    Arguments pertaining to what data we are going to input our model for training and eval.
    """

    dataset: str = field(
        metadata={"help": "Dataset dataset"}
    )
    validation_dataset: Optional[str] = field(
        default=None,
        metadata={"help": "Dataset dataset for validation set"}
    )
    reader: Optional[str] = field(
        default="base",
        metadata={"help": "Dataset dataset"}
    )
    validation_reader: Optional[str] = field(
        default=None,
        metadata={"help": "Dataset dataset"}
    )
    postprocessing: Optional[str] = field(
        default="default",
        metadata="post processing to do"
    )
    scorer: Optional[str] = field(
        default="classification",
        metadata="loading scorer function"
    )
    experiment_name: Optional[str] = field(
        default=None,
        metadata={"help": "Experiment name"}
    )
    out_file: Optional[str] = field(default=None, metadata={"help":"Save predictions to this file"})
    project_name: Optional[str] = field(
        default=None,
        metadata={"help": "Experiment project"}
    )
    train_file: Optional[str] = field(default=None, metadata={"help": "The input training data file (a text file)."})
    validation_file: Optional[str] = field(
        default=None,
        metadata={"help": "An optional input evaluation data file to evaluate the perplexity on (a text file)."},
    )
    test_file: Optional[str] = field(
        default=None,
        metadata={"help": "An optional input test data file to evaluate the perplexity on (a text file)."},
    )
    overwrite_cache: bool = field(
        default=False, metadata={"help": "Overwrite the cached training and evaluation sets"}
    )
    preprocessing_num_workers: Optional[int] = field(
        default=None,
        metadata={"help": "The number of processes to use for the preprocessing."},
    )
    max_seq_length: int = field(
        default=192,
        metadata={
            "help": "The maximum total input sequence length after tokenization. Sequences longer "
            "than this will be truncated, sequences shorter will be padded."
        },
    )
    pad_to_max_length: bool = field(
        default=True,
        metadata={
            "help": "Whether to pad all samples to `max_seq_length`. "
            "If False, will pad the samples dynamically when batching to the maximum length in the batch (which can "
            "be faster on GPU but will be slower on TPU)."
        },
    )
    max_train_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": "For debugging purposes or quicker training, truncate the number of training examples to this "
            "value if set."
        },
    )
    max_eval_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": "For debugging purposes or quicker training, truncate the number of evaluation examples to this "
            "value if set."
        },
    )
    max_predict_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": "For debugging purposes or quicker training, truncate the number of prediction examples to this "
            "value if set."
        },
    )
    version_2_with_negative: bool = field(
        default=False, metadata={"help": "If true, some of the examples do not have an answer."}
    )
    null_score_diff_threshold: float = field(
        default=0.0,
        metadata={
            "help": "The threshold used to select the null answer: if the best answer has a score that is less than "
            "the score of the null answer minus this threshold, the null answer is selected for this example. "
            "Only useful when `version_2_with_negative=True`."
        },
    )
    doc_stride: int = field(
        default=128,
        metadata={"help": "When splitting up a long document into chunks, how much stride to take between chunks."},
    )
    n_best_size: int = field(
        default=20,
        metadata={"help": "The total number of n-best predictions to generate when looking for an answer."},
    )
    max_target_length: int = field(
        default=30,
        metadata={
            "help": "The maximum length of an answer that can be generated. This is needed because the start "
            "and end predictions are not conditioned on one another."
        },
    )
    train_beam: int = field(
        default=1
    )
    eval_beam: int = field(
        default=10
    )
    ignore_pad_token_for_loss: bool = field(
        default=True,
        metadata={
            "help": "Whether to ignore the tokens corresponding to padded labels in the loss computation or not."
        },
    )
    package: Optional[str] = field(
        default=None,
        metadata={"help": "Package to import"}
    )


    def __post_init__(self):
        if (self.train_file is None
            and self.validation_file is None
            and self.test_file is None
        ):
            raise ValueError("Need either a dataset name or a training/validation file/test_file.")
        else:
            if self.train_file is not None:
                extension = self.train_file.split(".")[-1]
            if self.validation_file is not None:
                extension = self.validation_file.split(".")[-1]
            if self.test_file is not None:
                extension = self.test_file.split(".")[-1]
